# -*- coding: UTF-8 -*-
# pylint: disable=no-member, unused-argument


""" Additional SQL migrations for SQL databases. At the moment only mysql is
implemented. """


class ActionType:
    """ ON events action types """
    CASCADE = 0
    SET_NULL = 1
    PROTECT = 2
    NO_ACTION = 3


class EventType:
    """ ON event types """
    UPDATE = 0
    DELETE = 1

_mysql_actions = {
    ActionType.CASCADE: "CASCADE",
    ActionType.SET_NULL: "SET NULL",
    ActionType.PROTECT: "RESTRICT",
}


class CustomMigrationsBase(object):
    """ Base class for custom migrations """
    @classmethod
    def get_db_info(cls, schema_editor):
        """ Get database name and alias """
        from django.conf import settings
        alias = schema_editor.connection.alias
        dbname = settings.DATABASES[alias]['NAME']
        return (alias, dbname)


class AddDefaultValueBase(CustomMigrationsBase):
    """ Class migrating for default value fixes """
    @classmethod
    def change_func(cls, apps, schema_editor):
        """ Add timestamps """
        if schema_editor.connection.vendor != 'mysql':
            return
        for table, field, default_value in cls.fields:
            schema_editor.execute(
                """
                ALTER TABLE
                  `{0}`
                ALTER
                  `{1}`
                SET DEFAULT {2}
                """.format(
                    table,
                    field,
                    default_value
                )
            )

    @classmethod
    def drop_func(cls, apps, schema_editor):
        """ Add timestamps """
        if schema_editor.connection.vendor != 'mysql':
            return
        for table, field, _ in cls.fields:
            schema_editor.execute(
                """
                ALTER TABLE
                  `{0}`
                ALTER
                  `{1}`
                DROP DEFAULT
                """.format(
                    table,
                    field
                )
            )


class AddAutoTimestampBase(CustomMigrationsBase):
    """ Class migrating timestamp fixes """
    on_update = "ON UPDATE CURRENT_TIMESTAMP"

    @classmethod
    def forwards_func(cls, apps, schema_editor):
        """ Add timestamps """
        if schema_editor.connection.vendor != 'mysql':
            return
        for table, field, do_on_update  in cls.fields:
            on_upate_str = ""
            if do_on_update:
                on_upate_str = cls.on_update
            schema_editor.execute(
                """
                ALTER TABLE
                  `{0}`
                CHANGE
                  `{1}` `{1}`
                TIMESTAMP {2} NOT NULL
                DEFAULT CURRENT_TIMESTAMP
                """.format(
                    table,
                    field,
                    on_upate_str
                )
            )

    @classmethod
    def backwards_func(cls, apps, schema_editor):
        """ Remove timestamps """
        if schema_editor.connection.vendor != 'mysql':
            return
        for table, field, _ in cls.fields:
            schema_editor.execute(
                """
                ALTER TABLE
                  `{0}`
                CHANGE
                  `{1}` `{1}`
                DATETIME NOT NULL
                """.format(
                    table,
                    field
                )
            )


class AddOnEventsBase(CustomMigrationsBase):
    """ Class adding cascade fixes """
    @classmethod
    def change_func(cls, apps, schema_editor):
        """ Add events """
        cls.main_func(apps, schema_editor, False)

    @classmethod
    def delete_func(cls, apps, schema_editor):
        """ Remove events """
        cls.main_func(apps, schema_editor, True)

    @classmethod
    def main_func(cls, apps, schema_editor, delete):
        """ Add/remove events """
        # pylint: disable=too-many-locals
        _, dbname = cls.get_db_info(schema_editor)
        conn = schema_editor.connection
        if conn.vendor != 'mysql':
            return
        for table, field, commands in cls.fields:
            with conn.cursor() as cursor:
                query = """
                SELECT
                    `CONSTRAINT_NAME`,
                    `REFERENCED_TABLE_NAME`, 
                    `REFERENCED_COLUMN_NAME`
                FROM
                    `information_schema`.`KEY_COLUMN_USAGE`
                WHERE
                    `CONSTRAINT_SCHEMA` = '{database}' AND
                    `TABLE_NAME` = '{table}' AND
                    `COLUMN_NAME` = '{column}' AND
                    `REFERENCED_TABLE_NAME` IS NOT NULL;
                """.format(
                    database = dbname,
                    table = table,
                    column = field,
                )
                try:
                    cursor.execute(query)
                    (
                        fk_name,
                        ref_table,
                        ref_column,
                    ) = cursor.fetchone()
                except Exception:
                    print(query)
                    raise
            schema_editor.execute(
                """
                ALTER TABLE
                    `{table}`
                DROP FOREIGN KEY
                    `{fk_name}`;
                """.format(
                    table = table,
                    fk_name = fk_name,
                )
            )
            on_update = ""
            on_delete = ""
            if not delete:
                if EventType.UPDATE in commands:
                    on_update = "ON UPDATE {0}".format(
                        _mysql_actions[commands[EventType.UPDATE]]
                    )
                if EventType.DELETE in commands:
                    on_delete = "ON DELETE {0}".format(
                        _mysql_actions[commands[EventType.DELETE]]
                    )
            query = """
            ALTER TABLE
                `{table}`
            ADD CONSTRAINT
                `{fk_name}`
            FOREIGN KEY
                (`{column}`)
            REFERENCES
                `{ref_table}` (`{ref_column}`)
            {on_update}
            {on_delete};
            """.format(
                table      = table,
                fk_name    = fk_name,
                column     = field,
                ref_table  = ref_table,
                ref_column = ref_column,
                on_update  = on_update,
                on_delete  = on_delete,
            )
            try:
                schema_editor.execute(query)
            except Exception:
                print(query)
                raise
